from typing import Optional
from abc import ABC
from asyncio import iscoroutinefunction
from pydantic import BaseModel, ValidationError

try:
    from logzero import logger
except ImportError:
    import logging as logger


class Headers:
    def __init__(
        self,
        role=None,
        user_id=None,
        key=None,
        **kwargs,
    ):
        headers = {"x-gapo-role": role or "service"}

        if key:
            headers.update({"x-gapo-api-key": key})

        if user_id:
            headers.update({"x-gapo-user-id": user_id})

        if kwargs:
            headers.update(kwargs)

        self._h = headers

    def dict(self):
        return self._h


class AbstractRequest(ABC):
    service: str
    endpoint: str = ""
    key: str
    response: Optional
    Response: Optional
    Params: Optional

    def __init__(self, client_cfg: dict, client_session):
        self.url = client_cfg[self.service] + "/" + self.endpoint

        api_key = client_cfg.get(
            getattr(self, "key", None) or (self.service + "___KEY")
        )

        self.headers = Headers(
            role=getattr(self, "role", None),
            key=api_key,
            user_id=getattr(self, "user_id", None),
        )

        self._cl = client_session

    async def get(self, **params):
        try:
            if hasattr(self, "Params"):
                params = self.Params(**params).dict()

            headers = self.headers.dict()
            request_cfg = {"params": params, "headers": headers}

            async with self._cl.get(self.url, **request_cfg) as resp:
                if resp.status != 200:
                    msg = "Url=%s, API call unsuccesful => %s"
                    text = await resp.text()
                    logger.error(msg, self.url, text)
                    return self.on_failure(resp)

                json = await resp.json()

                if iscoroutinefunction(self.on_success):
                    return await self.on_success(json)

                return self.on_success(json)

        except ValidationError as err:
            logger.error("Url=%s, Invalid query-param: %s", self.url, err)
            return self.on_error(err)

        except Exception as err:
            logger.error("Url=%s, API call error: %s", self.url, err)
            return self.on_error(err)

    def on_failure(self, response):
        pass

    def on_error(self, error):
        pass

    def on_success(self, json_data):
        if hasattr(self, "Response"):
            return self.Response(**json_data)

        if hasattr(self, "response"):
            return self.response(**json_data)

        return json_data



""" Typical Linkedlist Node
"""
from gapoic.exceptions import (
    ReadonlyException,
    InvalidDataTypeException,
)


class Node:
    """Node instance can hold any kind of data
    - For doubly-linkedlist
    - By default, node's data is immutable
    """

    def __init__(self, key, data, mutable=False):
        self.__key = key
        self.__data = data
        self.__mutable = mutable
        self.__next: Node = None
        self.__prev: Node = None

    @property
    def data(self):
        return self.__data

    @data.setter
    def data(self, data):
        if not self.__mutable:
            raise ReadonlyException("Node's data")
        self.__data = data

    @property
    def key(self):
        return self.__key

    @key.setter
    def key(self, _):
        raise ReadonlyException("Node's key")

    @property
    def next(self):
        return self.__next

    @next.setter
    def next(self, some_node):
        if some_node is not None and not isinstance(some_node, self.__class__):
            raise InvalidDataTypeException(self.__class__)
        self.__next = some_node

    @property
    def prev(self):
        return self.__prev

    @prev.setter
    def prev(self, some_node):
        if some_node is not None and not isinstance(some_node, self.__class__):
            raise InvalidDataTypeException(self.__class__)
        self.__prev = some_node



""" ABS exceptions
"""


class ReadonlyException(Exception):
    """Raise when user is trying to set value on a readonly property"""

    def __init__(self, some_class_property):
        self.message = f"{some_class_property} is READ-ONLY"
        super().__init__(self.message)


class InvalidDataTypeException(Exception):
    """Raise when user is passing a value of non-acceptable data type"""

    def __init__(self, some_type):
        self.message = f"Only value of {some_type} is ACCEPTED"
        super().__init__(self.message)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gapoic', 'gapoic.abs', 'gapoic.http']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['cover = scripts:cover',
                     'lint = scripts:lint',
                     'test = scripts:test',
                     'test_gapo = scripts:test_gapo',
                     'test_misc = scripts:test_misc']}

setup_kwargs = {
    'name': 'gapoic',
    'version': '0.3.1',
    'description': 'pythonic tools we use at GAPO',
    'long_description': None,
    'author': 'vutr',
    'author_email': 'me@vutr.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Doubly-Linked-List"""
from typing import Optional
from gapoic.abs.node import Node
from gapoic.exceptions import InvalidDataTypeException


class DoublyLinkedList:
    """DoublyLinkedList"""

    def __init__(self, head: Node = None):
        self.__head = head

    def __str__(self):
        node: Optional[Node] = self.head
        text = ""

        while node:
            text += f"{node.key}"
            if node.next:
                text += " > "
            node = node.next

        return text

    @property
    def head(self):
        """return the head node of the linked-list"""
        return self.__head

    @head.setter
    def head(self, new_head: Optional[Node]):
        """change head, must be a Node"""
        if new_head is not None and not isinstance(new_head, Node):
            raise InvalidDataTypeException(Node)
        self.__head = new_head
